@@ -63,28 +63,28 @@ class BBOXNormalizedLine():
         self.boundingbox = BoundingBox
         self.text = ''
         self.merged = merged
-        self.__calculateMedians()
+        self.calculateMedians()
         self.__appendText(Text)
         self.avg_height = avg_height
         self.std_height = std_height
         self.rank=0.0
         self.listids=[]
 
-    def __calculateMedians(self):
+    def calculateMedians(self):
         self.xmedian=(min(self.boundingbox[0].X,self.boundingbox[3].X) + max(self.boundingbox[1].X,self.boundingbox[2].X))/2
         self.ymedian=(min(self.boundingbox[0].Y,self.boundingbox[3].Y) + max(self.boundingbox[1].Y,self.boundingbox[2].Y))/2
 
     def getClusterId(self):
         return self.listids[0]
 
-    def getBoxesAsArray(self):
+    def getBoxesAsArray(self,scale=1):
         result=[]
         for box in self.boundingbox:
-            result.append([box.X,box.Y])
+            result.append([box.X*scale,box.Y*scale])
         return result
 
-    def getBoxesAsRectangle(self):
-        return (self.boundingbox[0].X,self.boundingbox[0].Y,(self.boundingbox[2].X-self.boundingbox[0].X),(self.boundingbox[2].Y-self.boundingbox[0].Y))
+    def getBoxesAsRectangle(self,scale=1):
+        return (self.boundingbox[0].X*scale,self.boundingbox[0].Y*scale,(self.boundingbox[2].X-self.boundingbox[0].X)*scale,(self.boundingbox[2].Y-self.boundingbox[0].Y)*scale)
 
     def __appendText(self, Text):
         self.text += Text
@@ -100,41 +100,43 @@ class BBOXNormalizedLine():
         self.boundingbox[1] = BBoxUtils.maxXminY(1,self,line)
         self.boundingbox[2] = BBoxUtils.maxXmaxY(2,self,line)
         self.boundingbox[3] = BBoxUtils.minXmaxY(3,self,line)
-        self.__calculateMedians()
+        self.calculateMedians()
         self.end_idx=line.end_idx
         
     @classmethod
-    def from_azure(cls, index, data):
+    def from_azure(cls, index, data, ppi):
         points = list()
         array=data["boundingBox"]
         if ( len(array) > 4 ):
             x = array[0]
             y = array[1]          
-            points.append(BBOXPoint(x,y))
+            points.append(BBOXPoint(x*ppi,y*ppi))
             x = array[2]
             y = array[3]
-            points.append(BBOXPoint(x, y))
+            points.append(BBOXPoint(x*ppi, y*ppi))
             x = array[4]
             y = array[5]
-            points.append(BBOXPoint(x, y))
+            points.append(BBOXPoint(x*ppi, y*ppi))
             x = array[6]
             y = array[7]
-            points.append(BBOXPoint(x, y))
+            points.append(BBOXPoint(x*ppi, y*ppi))
+
             # Make sure the X of the line is consistent to the first word of the same. 
             if len(data['words'])>0:
                 points[0].X = data['words'][0]["boundingBox"][0]+1
                 points[3].X = data['words'][0]["boundingBox"][6]+1
         else:
-            points = list(map(BBOXPoint.from_azure, array))
+            points = list(map(BBOXPoint.from_azure, [ppi*x for x in array]))
             if len(data['words'])>0:
                 points[0].X = data['words'][0]["boundingBox"][0].X+1
                 points[3].X = data['words'][0]["boundingBox"][3].X+1
           
         wordheights=[]
-        # Check Line in anomaly
+        # # Check Line in anomaly
         for wordbox in data['words']:
             wordheights.append(wordbox['boundingBox'][5]-wordbox['boundingBox'][1])
-        # calculate the average 
+
+        # calculate the average
         npheights = np.array(wordheights)
         avg_height = np.average(npheights)
         std_height = np.std(npheights, dtype=np.float64)
@@ -159,7 +161,7 @@ class BBOXNormalizedLine():
         return BBOXNormalizedLine(Idx=line_counter,BoundingBox=points,Text=line_text)
 
 class BBOXPageLayout():
-    def __init__(self, Id:int = 0,clockwiseorientation:float = 0.0,Width:float = 0.0,Height:float = 0.0,Unit:str = "pixel",Language:str="en",Text:str=None,Lines:List[BBOXNormalizedLine]=None):
+    def __init__(self, Id:int = 0,clockwiseorientation:float = 0.0,Width:float = 0.0,Height:float = 0.0,Unit:str = "pixel",Language:str="en",Text:str=None,Lines:List[BBOXNormalizedLine]=None,ppi=1):
         self.id=Id
         self.clockwiseorientation=clockwiseorientation
         self.width=Width
@@ -168,16 +170,28 @@ class BBOXPageLayout():
         self.language=Language
         self.text=Text
         self.lines=Lines
+        self.ppi=ppi
     @classmethod
     def from_azure(cls, data):
-        lines=[BBOXNormalizedLine.from_azure(i,line) for i,line in enumerate(data["lines"])] 
+        if data["unit"]=="inch":
+            # Azure OCR response doesn't provide the ppi per page
+            # so we need to determine it for normalizing the processing of lines
+            ppi=BBoxUtils.determine_ppi(data["width"],data["height"])
+            # decimal precision on Azure is set to 4, so we can set a 10000 to normalize the box and 
+            # not convert to pixel
+            ppi=10000
+        else:
+            ppi=1
+
+        lines=[BBOXNormalizedLine.from_azure(i,line, 1) for i,line in enumerate(data["lines"])] 
         # >=0.6.0
         if "angle" in data:
             angle = data["angle"]
         # <= 0.5.0
         elif "clockwiseorientation" in data:
             angle = data["clockwiseorientation"]
-        return cls(Id=data["page"],clockwiseorientation=angle,Width=data["width"],Height=data["height"],Unit=data["unit"],Lines=lines)
+          
+        return cls(Id=data["page"],clockwiseorientation=angle,Width=data["width"],Height=data["height"],Unit=data["unit"],Lines=lines,ppi=ppi)
 
     @classmethod
     def from_google(cls, page):
@@ -253,6 +267,7 @@ class BBOXOCRResponse():
         self.original_text=original_text
         self.text=Text
         self.pages=recognitionResults
+
     @classmethod
     def from_azure(cls, data):
         # Convert to JSON dict if a JSON string is passed.
@@ -317,34 +332,72 @@ class BBoxHelper():
     def __processOCRResponse(self, response, sortingAlgo=BBoxSort.contoursSort, boxSeparator:str = None):
         """processOCRResponse method
         Process an OCR Response input (Azure,Google) and returns a new BBox format OCR response.
+
+        Iterate through each page of the response and process it. Before processing we tackle the boxes 
+        orientation. 
         """
         newtext = ""
+
         # Rotate the BBox of each page based on its corresponding orientation
-        for item in response.pages:
-            bboxlogger.debug("Processing Page {}".format(str(item.id)))
+        for page in response.pages:
+            bboxlogger.debug("Processing Page {0} with {1} lines.".format(str(page.id),len(page.lines)))
             if self.annotate and bboxannotate.pageTag:
                 newtext+=bboxannotate.pageTag[0]
-            rotation = round(item.clockwiseorientation,0)
+            rotation = round(page.clockwiseorientation,0)
             bboxlogger.debug("Orientation {}".format(str(rotation)))
-            # TODO Do the Math for clockwise orientation
-            if (rotation >= 360-1 or rotation == 0):
+            page_width=page.width
+            page_height=page.height
+            # Bounding Boxes Orientation
+            # if ( rotation != 0):
+            #     # TODO Do the Math for clockwise orientation
+            #     for line in page.lines:
+            #         bboxlogger.debug("Before rotation {}".format(line.boundingbox))
+            #         test = BBoxUtils.rotateLineBoundingBox(line.boundingbox, -1*page.clockwiseorientation)
+            #         bboxlogger.debug("After rotation 1 {}".format(test))
+            #         test = BBoxUtils.rotateLineBoundingBox(line.boundingbox, page.clockwiseorientation)
+            #         bboxlogger.debug("After rotation 2 {}".format(test))
+
+            rotation_threshold=1
+
+            applied_rotation=0 
+            # if (rotation >= 360-1 or rotation == 0):
+            if rotation in range(360-rotation_threshold,360+rotation_threshold) or rotation in range(-rotation_threshold,rotation_threshold):
                 bboxlogger.debug("no rotation adjustment required")
-            elif (rotation >= 270-1):
+            # elif (rotation >= 270-1) or rotation in range(-180,-90):
+            elif rotation in range(270-rotation_threshold,270+rotation_threshold) or rotation in range(-90-rotation_threshold,-90+rotation_threshold):
+                applied_rotation=90
                 # //Rotate 90 clockwise
-                for x in item.lines:
-                    x.boundingbox = BBoxUtils.rotateBoundingBox(item.width, item.height, x.boundingbox, -90)
-            elif (rotation >= 180-1):
+                for line in page.lines:
+                    line.boundingbox = BBoxUtils.rotateBoundingBox(page.width, page.height, line.boundingbox, applied_rotation)
+                    line.calculateMedians()
+                # Switch W/H accordingly
+                page.width = page_height
+                page.height = page_width
+            # elif (rotation >= 180-1):
+            elif rotation in range (180-rotation_threshold,180+rotation_threshold) or rotation in range(-180-rotation_threshold,-180+rotation_threshold):
+                applied_rotation=180
                 # // Rotate 180
-                for x in item.lines:
-                    x.boundingbox = BBoxUtils.rotateBoundingBox(item.width, item.height, x.boundingbox, 180)
-            elif (rotation >= 90-1):
+                for line in page.lines:
+                    line.boundingbox = BBoxUtils.rotateBoundingBox(page.width, page.height, line.boundingbox, applied_rotation)
+                    line.calculateMedians()
+            # elif (rotation >= 90-1):
+            elif rotation in range(90-rotation_threshold,90+rotation_threshold) or rotation in range(-270-rotation_threshold,-270+rotation_threshold):
+                applied_rotation=-90
                 # //Rotate 90 counterclockwise
-                for x in item.lines:
-                    x.boundingbox = BBoxUtils.rotateBoundingBox(item.width, item.height, x.boundingbox, 90)
+                for line in page.lines:
+                    line.boundingbox = BBoxUtils.rotateBoundingBox(page.width, page.height, line.boundingbox, applied_rotation)
+                    line.calculateMedians()
+                # Switch W/H accordingly
+                page.width = page_height
+                page.height = page_width
             else:
                 bboxlogger.info("TODO rotation adjustment required ? ")
 
-            newtext += self.processOCRPageLayout(item, sortingAlgo, boxSeparator).text
+            # Invoke the page processing
+            page = self.processOCRPageLayout(page, sortingAlgo, boxSeparator)
+            newtext += page.text
+
+            bboxlogger.debug("Processed Page {0} with {1} bbox lines.".format(str(page.id),len(page.lines)))
 
             if self.annotate and bboxannotate.pageTag:
                 newtext+=bboxannotate.pageTag[1]
@@ -352,20 +405,30 @@ class BBoxHelper():
                 # Default page separator
                 newtext += '\r\n'
 
+            # Revert the rotation of the bounding boxes back to its original orientation
+            if applied_rotation!=0:
+                for line in page.lines:
+                    line.boundingbox = BBoxUtils.rotateBoundingBox(page.width, page.height, line.boundingbox, int(-applied_rotation))
+                    # recalculate medians for drawing the boxes correctly
+                    line.calculateMedians()
+                # Restore W/H 
+                page.width = page_width
+                page.height = page_height
+
         response.text = str(newtext)
 
         return response
 
-    def __processLineBoundingBoxes(self, lines, alignment, unit): 
+    def __processLineBoundingBoxes(self, lines, alignment, unit, ppi): 
         boxref = 0
-        Xthresholdratio = bboxconfig.config[unit].Thresholds[alignment].Xthresholdratio
-        Ythresholdratio = bboxconfig.config[unit].Thresholds[alignment].Ythresholdratio
+        Xthresholdratio = bboxconfig.get_Thresholds(unit,ppi)[alignment].Xthresholdratio
+        Ythresholdratio = bboxconfig.get_Thresholds(unit,ppi)[alignment].Ythresholdratio
         if ( alignment == LeftAlignment):
             boxref = 0
             # //Adjustment for bullet list text. 
             for line in lines:
                 if (line.text.startswith(". ")):
-                    line.boundingbox[boxref].X = line.boundingbox[boxref].X + bboxconfig.config[unit].ImageTextBoxingBulletListAdjustment
+                    line.boundingbox[boxref].X = line.boundingbox[boxref].X + bboxconfig.get_ImageTextBoxingBulletListAdjustment(unit,ppi)
             XSortedList = sorted([o for o in lines if o.merged == False],key= lambda o: (o.boundingbox[boxref].X,o.boundingbox[boxref].Y))
         elif (alignment == RightAlignment):
             boxref = 1
@@ -390,8 +453,8 @@ class BBoxHelper():
             if (alignment == CenteredAlignment):
                 xcurrent = line.xmedian
 
-            lowb=(regionx - (Xthresholdratio * bboxconfig.config[unit].ImageTextBoxingXThreshold))
-            highb=(regionx + (Xthresholdratio * bboxconfig.config[unit].ImageTextBoxingXThreshold))
+            lowb=(regionx - (Xthresholdratio * bboxconfig.get_ImageTextBoxingXThreshold(unit,ppi)))
+            highb=(regionx + (Xthresholdratio * bboxconfig.get_ImageTextBoxingXThreshold(unit,ppi)))
             if lowb>0.0:
                 bboxlogger.debug("{6}|Region Id:{0} X:{1} LowX {2}<{3}<{4} HighX | Same Region? {5} ".format(str(regionidx),str(regionx),str(lowb),str(xcurrent),str(highb),str((xcurrent >= lowb and xcurrent <= highb)),alignment))
 
@@ -425,7 +488,7 @@ class BBoxHelper():
                 ycurrent = line.boundingbox[boxref].Y
                 # lowb=(regiony - (Ythresholdratio * bboxconfig.config[unit].ImageTextBoxingYThreshold))
                 lowb=regiony-(regiony*0.1)
-                highb=(regiony + (Ythresholdratio * bboxconfig.config[unit].ImageTextBoxingYThreshold))
+                highb=(regiony + (Ythresholdratio * bboxconfig.get_ImageTextBoxingYThreshold(unit,ppi)))
                 bboxlogger.debug("{7}|Line bbox {0} {6} {1} | LowY {2}<{3}<{4} HighY | Merge {5}".format(str(line.boundingbox),str(line.text),str(lowb),str(ycurrent),str(highb),str((ycurrent >= lowb and ycurrent < highb)),str(regiony),alignment))
 
                 if (regiony == 0.0):
@@ -455,10 +518,13 @@ class BBoxHelper():
         # Go through potential Text Alignment : Left, Right and Centered.
         for alignment in Alignments:
             bboxlogger.debug("{1}|Processing {0}".format(alignment,str(page.id)))
-            page.lines = self.__processLineBoundingBoxes(page.lines,alignment,page.unit)
+            page.lines = self.__processLineBoundingBoxes(page.lines,alignment,page.unit,page.ppi)
 
         outlines=[o for o in page.lines if o.merged == False]
-        bboxlogger.debug("{1}|Output # lines {0}".format(len(outlines),str(page.id)))
+        bboxlogger.debug("{1}|Output {0} lines before sorting...".format(len(outlines),str(page.id)))
+        #
+        # Page lines Sorting
+        # 
 
         if sortingAlgo is None:
             # Old Default Sorting Strategy 
@@ -475,7 +541,9 @@ class BBoxHelper():
                 bboxlogger.debug("{0}|Default sorting strategy".format(str(page.id)))
                 sortedBlocks = sorted(outlines,key= lambda o: (o.boundingbox[0].X, o.ymedian))
         else:
-            sortedBlocks = sortingAlgo(page.id,page.width,page.height,blocks=outlines)
+            sortedBlocks = sortingAlgo(page.id,page.width,page.height,blocks=outlines,scale=page.ppi)
+
+        bboxlogger.debug("{1}|Output {0} lines after  sorting...".format(len(sortedBlocks),str(page.id)))
 
         # Output the actual from the sorted blocks
         newtext = ""
@@ -519,4 +587,5 @@ class BBoxHelper():
         page.lines = sortedBlocks
         # // Updating the Text after our processing.
         page.text = newtext
+
         return page
